#ifndef _ThreadPool_h_
#define _ThreadPool_h_

//! @file
//!     Includes boost asio thread_pool headers, or defines a dummy implementation.

#include <boost/asio/thread_pool.hpp>
#include <boost/asio/post.hpp>

#endif
