//! GiGi - A GUI for OpenGL
//!
//!  Copyright (C) 2017 Marcel Metz
//!  Copyright (C) 2018-2020 The FreeOrion Project
//!
//! Released under the GNU Lesser General Public License 2.1 or later.
//! Some Rights Reserved.  See COPYING file or https://www.gnu.org/licenses/lgpl-2.1.txt
//! SPDX-License-Identifier: LGPL-2.1-or-later

#define BOOST_TEST_MODULE "GG unit/coverage tests"
#include <boost/test/unit_test.hpp>
