//! GiGi - A GUI for OpenGL
//!
//!  Copyright (C) 2003-2008 T. Zachary Laine <whatwasthataddress@gmail.com>
//!  Copyright (C) 2013-2020 The FreeOrion Project
//!
//! Released under the GNU Lesser General Public License 2.1 or later.
//! Some Rights Reserved.  See COPYING file or https://www.gnu.org/licenses/lgpl-2.1.txt
//! SPDX-License-Identifier: LGPL-2.1-or-later

//! @file GG/Config.h
//!
//! Contains build-generated configuration macros used throughout GG.

#ifndef _GG_Config_h_
#define _GG_Config_h_


#define GG_HAVE_LIBPNG 1
#define GG_HAVE_LIBTIFF 0


#endif
