#ifndef PYTHON_WRAPPERS
#define PYTHON_WRAPPERS


namespace FreeOrionPython {
    void WrapUniverseClasses();
    void WrapGalaxySetupData();
    void WrapGameStateEnums();
    void WrapEmpire();
    void WrapLogger();
    void WrapConfig();
}

#endif
