#ifndef _EnumsFwd_h_
#define _EnumsFwd_h_

#include <stdint.h>

enum class UniverseObjectType : int8_t;
enum class StarType : int8_t;
enum class PlanetType : int8_t;
enum class PlanetSize : int8_t;
enum class PlanetEnvironment : int8_t;
enum class MeterType : int8_t;
enum class EmpireAffiliationType : int8_t;
enum class UnlockableItemType : int8_t;
enum class ResourceType : int8_t;
enum class ShipPartClass : int8_t;
enum class ShipSlotType : int8_t;
enum class Visibility : int8_t;
enum class CaptureResult : int8_t;
enum class ModeratorActionSetting : int8_t;
enum class FleetAggression : int8_t;
enum class DiplomaticStatus : int8_t;

#endif
