//! GiGi - A GUI for OpenGL
//!
//!  Copyright (C) 2015 Marcel Metz
//!  Copyright (C) 2016-2020 The FreeOrion Project
//!
//! Released under the GNU Lesser General Public License 2.1 or later.
//! Some Rights Reserved.  See COPYING file or https://www.gnu.org/licenses/lgpl-2.1.txt
//! SPDX-License-Identifier: LGPL-2.1-or-later

/** \file FontFwd.h \brief Contains forward declaration of all public
    GG control and utility classes. */

#ifndef _GG_GGFwd_h_
#define _GG_GGFwd_h_


namespace GG {

class Button;
class DropDownList;
class DynamicGraphic;
class Edit;
class Font;
class ListBox;
class MultiEdit;
class RadioButtonGroup;
class Scroll;
class ScrollPanel;
class StaticGraphic;
class SubTexture;
class TabWnd;
class Texture;
class TextControl;
class RichText;
struct Clr;

typedef TextControl Label;

template <typename T>
class Spin;

}


#endif
